/*
Copyright (c) 2004 Paul Hsieh
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

    Neither the name of sorttest nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

/*

Recommended flags:
------------------

Intel C/C++:
icl /O2 /G6 /Qaxi /Qxi /Qip sorttest.c

WATCOM C/C++:
wcl386 /otexan /6r sorttest.c

GCC:
gcc -O3 -mcpu=athlon-xp -march=athlon-xp sorttest.c

MSVC:
cl /O2 /Ot /Og /G6 sorttest.c

*/

/*
 *  Direct small sort functions
 */

static void sort2 (int * numbers) {
int tmp;

    if (numbers[0] <= numbers[1]) return;
    tmp = numbers[0];
    numbers[0] = numbers[1];
    numbers[1] = tmp;
}

static void sort3 (int * numbers) {
int tmp;

    if (numbers[0] <= numbers[1]) {
        if (numbers[1] <= numbers[2]) return;
        if (numbers[2] <= numbers[0]) {
            tmp = numbers[0];
            numbers[0] = numbers[2];
            numbers[2] = numbers[1];
            numbers[1] = tmp;
            return;
        }
        tmp = numbers[1];
    } else {
        tmp = numbers[0];
        if (numbers[0] <= numbers[2]) {
            numbers[0] = numbers[1];
            numbers[1] = tmp;
            return;
        }
        if (numbers[2] <= numbers[1]) {
            numbers[0] = numbers[2];
            numbers[2] = tmp;
            return;
        }
        numbers[0] = numbers[1];
    }
    numbers[1] = numbers[2];
    numbers[2] = tmp;
}

static void sortAlt2 (int * numbers, int * altNumbers) {
    if (numbers[0] <= numbers[1]) {
        altNumbers[0] = numbers[0];
        altNumbers[1] = numbers[1];
    } else {
        altNumbers[0] = numbers[1];
        altNumbers[1] = numbers[0];
    }
}

static void sortAlt3 (int * numbers, int * altNumbers) {
    if (numbers[0] <= numbers[1]) {
        if (numbers[1] <= numbers[2]) {
            altNumbers[0] = numbers[0];
            altNumbers[1] = numbers[1];
            altNumbers[2] = numbers[2];
        } else if (numbers[2] <= numbers[0]) {
            altNumbers[0] = numbers[2];
            altNumbers[1] = numbers[0];
            altNumbers[2] = numbers[1];
        } else {
            altNumbers[0] = numbers[0];
            altNumbers[1] = numbers[2];
            altNumbers[2] = numbers[1];
        }
    } else {
        if (numbers[0] <= numbers[2]) {
            altNumbers[0] = numbers[1];
            altNumbers[1] = numbers[0];
            altNumbers[2] = numbers[2];
        } else if (numbers[2] <= numbers[1]) {
            altNumbers[0] = numbers[2];
            altNumbers[1] = numbers[1];
            altNumbers[2] = numbers[0];
        } else {
            altNumbers[0] = numbers[1];
            altNumbers[1] = numbers[2];
            altNumbers[2] = numbers[0];
        }
    }
}

/*
 *  Heap Sort
 */

/* Assure the heap property for entries from top to last */
static void siftDown (int numbers[], int top, int last) {
int tmp = numbers[top];
int maxIdx = top;

/*
    while (last > (maxIdx += maxIdx)) {
        if (numbers[maxIdx] < numbers[maxIdx + 1]) maxIdx++;
        if (tmp >= numbers[maxIdx]) break;
        numbers[top] = numbers[maxIdx];
        top = maxIdx;
    }
    if (last == maxIdx && tmp < numbers[maxIdx]) {
        numbers[top] = numbers[maxIdx];
        top = maxIdx;
    }
    numbers[top] = tmp;
*/
    while (last >= (maxIdx += maxIdx)) {
        if (maxIdx != last && numbers[maxIdx] < numbers[maxIdx + 1]) maxIdx++;
        if (tmp >= numbers[maxIdx]) break;
        numbers[top] = numbers[maxIdx];
        top = maxIdx;
    }
    numbers[top] = tmp;
}

static void siftDown0 (int numbers[], int last) {
int tmp;

    if (numbers[0] < numbers[1]) {
        tmp = numbers[1];
        numbers[1] = numbers[0];
        siftDown (numbers, 1, last);
    } else {
        tmp = numbers[0];
    }
    numbers[0] = numbers[last];
    numbers[last] = tmp;
}

void heapSort (int numbers[], int qty) {
int i;

    if (qty <= 3) {
        if (qty == 3) sort3 (numbers);
        else if (qty == 2) sort2 (numbers);
        return;
    }

    i = qty / 2;
    /* Enforce the heap property for each position in the tree */
    for (  qty--; i >  0; i--) siftDown  (numbers, i, qty);
    for (i = qty; i >= 3; i--) siftDown0 (numbers, i);
    sort3 (numbers);
}

/*
 *  Quick Sort
 */

static int medianOf3 (int * numbers) {
int tmp, i = 2;

    if (numbers[0] <= numbers[1]) {
        if (numbers[2] <= numbers[0]) return numbers[0];
        if (numbers[1] <= numbers[2]) i = 1;
    } else {
        if (numbers[0] <= numbers[2]) return numbers[0];
        if (numbers[2] <= numbers[1]) i = 1;
    }
    tmp = numbers[i];
    numbers[i] = numbers[0];
    numbers[0] = tmp;
    return tmp;
}

static void quickSortRecurse (int numbers[], int left, int right) {
int pivot, lTmp, rTmp;

    if (right == left + 1) {
        sort2 (numbers + left);
        return;
    }
    if (right == left + 2) {
        sort3 (numbers + left);
        return;
    }
    if (right <= left) return;

    lTmp = left;
    rTmp = right;

    pivot = medianOf3 (numbers + left);

    goto QStart;
    while (1) {
        do {
            right--;
            if (left >= right) goto QEnd;
            QStart:;
        } while (numbers[right] > pivot);
        numbers[left] = numbers[right];
        do {
            left++;
            if (left >= right) {
                left = right;
                goto QEnd;
            }
        } while (numbers[ left] < pivot);
        numbers[right] = numbers[left];
    }
    QEnd:;
    numbers[left] = pivot;
    if (lTmp < left) quickSortRecurse (numbers,   lTmp, left-1);
    if (rTmp > left) quickSortRecurse (numbers, left+1,   rTmp);
}

void quickSort (int numbers[], int qty) {
    if (qty < 2) return;
    quickSortRecurse (numbers, 0, qty - 1);
}

/*
 *  Quick Sort 2
 */

static void insertionSort (int numbers[], int n) {
int i;

    for (i=1; i < n; i++)
    {
        int temp = numbers[i];
        int j = i;

        while (j > 0)
        {
            if (numbers[j-1] <= temp) break;

            numbers[j] = numbers[j-1];

            j--;
        }

        numbers[j] = temp;
    }
}

static void quickSortRecurse2 (int numbers[], int left, int right) {
int pivot, lTmp, rTmp;

    if (left + 20 > right) {
        if (right - left > 0) insertionSort (numbers + left, right - left + 1);
        return;
    }

    lTmp = left;
    rTmp = right;

    pivot = medianOf3 (numbers + left);

    goto QStart;
    while (1) {
        do {
            right--;
            if (left >= right) goto QEnd;
            QStart:;
        } while (numbers[right] > pivot);
        numbers[left] = numbers[right];
        do {
            left++;
            if (left >= right) {
                left = right;
                goto QEnd;
            }
        } while (numbers[ left] < pivot);
        numbers[right] = numbers[left];
    }

    QEnd:;
    numbers[left] = pivot;

    if (lTmp < left) quickSortRecurse2 (numbers,   lTmp, left-1);
    if (rTmp > left) quickSortRecurse2 (numbers, left+1,   rTmp);
}

void quickSort2 (int numbers[], int qty) {
    if (qty < 2) return;
    quickSortRecurse2 (numbers, 0, qty - 1);
}

/*
 *  Merge Sort
 */

static void mergesortInPlace (int * numbers, int * altNumbers, int qty);

/* Perform mergesort, but store results in altNumbers */

static void mergesortExchange (int * numbers, int * altNumbers, int qty) {
int half, i0, i1, i;

    if (qty == 2) {
        sortAlt2 (numbers, altNumbers);
        return;
    }
    if (qty == 3) {
        sortAlt3 (numbers, altNumbers);
        return;
    }

    half = (qty + 1)/2;

    mergesortInPlace (numbers, altNumbers, half);
    mergesortInPlace (numbers + half, altNumbers, qty - half);

    i0 = 0; i1 = half;

    for (i=0; i < qty; i++) {
        if (i1 >= qty || (i0 < half && numbers[i0] < numbers[i1])) {
            altNumbers[i] = numbers[i0];
            i0++;
        } else {
            altNumbers[i] = numbers[i1];
            i1++;
        }
    }
}

/* Perform mergesort and store results in numbers */

static void mergesortInPlace (int * numbers, int * altNumbers, int qty) {
int half, i0, i1, i;

    if (qty == 2) {
        sort2 (numbers);
        return;
    }
    if (qty == 3) {
        sort3 (numbers);
        return;
    }

    half = (qty + 1)/2;

    mergesortExchange (numbers, altNumbers, half);
    mergesortExchange (numbers + half, altNumbers + half, qty - half);

    i0 = 0; i1 = half;

    for (i=0; i < qty; i++) {
        if (i1 >= qty || (i0 < half && altNumbers[i0] < altNumbers[i1])) {
            numbers[i] = altNumbers[i0];
            i0++;
        } else {
            numbers[i] = altNumbers[i1];
            i1++;
        }
    }
}

#include <stdlib.h>

void mergeSort (int numbers[], int qty) {
int * tmpArray;

    if (qty <= 3) {
        if (qty == 3) sort3 (numbers);
        else if (qty == 2) sort2 (numbers);
        return;
    }

    tmpArray = (int *) malloc (qty * sizeof (int));
    mergesortInPlace (numbers, tmpArray, qty);
    free (tmpArray);
}

extern void __stdcall nrQsortA(int Arr[], int count);

void masm32_qsort (int numbers[], int qty) {

    nrQsortA(numbers, qty);
}

extern void __stdcall qisort(int arr[], int i, int j);

void steve_qisort (int numbers[], int qty) {

    qisort(numbers, 0, qty - 1);
}

#include <stdio.h>
#include <time.h>

#define NUM_ITEMS (600000u)
int numbers[NUM_ITEMS];

/* #define CORRECTNESS 1 */

unsigned int test (int seed, void (* sortfn)(int numbers[], int qty)) {
#ifdef CORRECTNESS
    int t0, t1, t2, t3, t4, t5, t6, t7;
#endif
    int i, c, n;
    clock_t clocks;
    clock_t total_clocks = 0;

    srand (seed);

    for (c = 0; c < 25; c++) {
#ifdef CORRECTNESS
        t0 = t1 = t2 = t3 = t4 = t5 = t6 = t7 = 0;
#endif

        n = NUM_ITEMS;

        /* fill array with random integers */
        for (i = 0; i < n; i++) {
            /* if (0 != (i & 15)) numbers[i] = (i + 1) % n;
            else */
            numbers[i] = rand ()*RAND_MAX + rand ();
#ifdef CORRECTNESS
            t0 += numbers[i];
            t1 ^= numbers[i];
            t2 += numbers[i] * numbers[i];
            t3 ^= numbers[i] * numbers[i];
#endif
        }

        /* perform heap sort on array */
        clocks = clock();
        sortfn (numbers, n);
        total_clocks += clock() - clocks;

#ifdef CORRECTNESS
        for (i = 0; i < n; i++) {
            if (i > 0 && numbers[i] < numbers[i-1]) {
                printf ("Error!  Not properly sorted! n = %d (%d > %d)\n", n, numbers[i-1], numbers[i]);
                return -1;
            }
            t4 += numbers[i];
            t5 ^= numbers[i];
            t6 += numbers[i] * numbers[i];
            t7 ^= numbers[i] * numbers[i];
        }

        if (t0 != t4 || t1 != t5 || t2 != t6 || t3 != t7) {
            printf ("The numbers are corrupted!\n");
            return -1;
        }
#endif
    }
    return total_clocks;
}

int main () {
int seed = 0; /* time (NULL) ^ getpid (); */
clock_t t0, t1, t2, t3, t4, t5, tmin;
time_t t;

    t = time (NULL);
    while (t == time (NULL));

    t0 = test (seed, heapSort);
    t1 = test (seed, quickSort);
    t2 = test (seed, mergeSort);
    t3 = test (seed, quickSort2);
    t4 = test (seed, masm32_qsort);
    t5 = test (seed, steve_qisort);

    tmin = (  t0 < t1) ?   t0 : t1;
    tmin = (tmin < t2) ? tmin : t2;
    tmin = (tmin < t3) ? tmin : t3;
    tmin = (tmin < t4) ? tmin : t4;
    tmin = (tmin < t5) ? tmin : t5;

    printf (" heapsort : %8u  %u%%\n", t0, 100*t0/tmin);
    printf ("quicksort : %8u  %u%%\n", t1, 100*t1/tmin);
    printf ("mergesort : %8u  %u%%\n", t2, 100*t2/tmin);
    printf ("qsort+ins : %8u  %u%%\n", t3, 100*t3/tmin);
    printf ("   masm32 : %8u  %u%%\n", t4, 100*t4/tmin);
    printf ("    steve : %8u  %u%%\n", t5, 100*t5/tmin);

    return 0;
}
